/* measureth.c : SHT-11制御 温度/湿度測定 サンプルプログラム
 *
 * コンパイル例:
 *	gcc -o measureth measureth.c -lrt
 *  (clock_gettime関数を使用するためリアルタイム拡張オプションが必要)
 *
 * 実行例：
 *	sudo ./measureth で実行. 温度/湿度を [Tmp:27.39, Hum:43.12]のように標準出力に返します.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdint.h>

/* Parameters & macros for RPI GPIO Control --------------------------------*/
// for mmap
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

// GPIO setup macros.
#define INP_GPIO_BIT(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))		//set 0b000
#define OUT_GPIO_BIT(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3)); \
						*(gpio+((g)/10)) |=  (1<<(((g)%10)*3))		//set 0b0000 -> 0b001
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

// GPIO Word Access
#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0
#define GPIO_GET *(gpio+13) // read   bits

// GPIO Bit Access
#define GPIO_SET_BIT(g) *(gpio+7) = 1<<g
#define GPIO_CLR_BIT(g) *(gpio+10) = 1<<g
#define GPIO_GET_BIT(g) (*(gpio+13) & (1<<g))

/* Parameters & macros for SHT1x device I/O Control -----------------------*/
// Assigned RPI GPIOpinNo	
#define RPI_SHTSCK 	3
#define RPI_SHTSDA 	2


#define SDA_IN() 	INP_GPIO_BIT(RPI_SHTSDA); waitUsec(2)
#define SDA_H() 	SDA_IN()
#define SDA_L() 	GPIO_CLR_BIT(RPI_SHTSDA); \
					OUT_GPIO_BIT(RPI_SHTSDA); waitUsec(2)
#define SCK_OUT() 	OUT_GPIO_BIT(RPI_SHTSCK)
#define SCK_H() 	GPIO_SET_BIT(RPI_SHTSCK); waitUsec(2)
#define SCK_L() 	GPIO_CLR_BIT(RPI_SHTSCK); waitUsec(2)
#define SCK_TRG() 	SCK_H(); SCK_L()

#define SDA_GET() 	GPIO_GET_BIT(RPI_SHTSDA)

// Parameters 
#define ACK 0
#define NAK 1
#define CMD_Temp 0b00000011		//Measure Temperature COMMAND
#define CMD_Hum  0b00000101		//Measure Relative Humidity COMMAND
#define CMD_SRES 0b00011110		//Soft reset COMMAND

uint8_t	crcval;
uint8_t sht_status_reg = 0;
char *gpio_map;

// I/O access for grobal
volatile unsigned *gpio;

//プロトタイプ宣言
void map_io();
void SHT_Init();
void SHT_Conreset();
void SHT_Trans_st();
void SHT_CalcCRC(uint8_t data);
uint8_t SHT_ReverseData(uint8_t);
uint8_t SHT_Write(uint8_t);
uint8_t SHT_Read(uint8_t);
uint8_t SHT_WaitforReady();
void waitUsec (uint32_t);
void sleepMsec (uint32_t);
void SHT_measureValue();


//------------------------------------------------------
// Set up a memory regions to access GPIO
//
void map_io(){		//GPIOレジスタをマッピング ポインタを戻す
	int mem_fd;
	
   /* open /dev/mem */
   if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
      printf("can't open /dev/mem \n");
      exit(-1);
   }

   /* mmap GPIO */
   gpio_map = (char *)mmap(
      NULL,             	//Any adddress in our space will do
      BLOCK_SIZE,       	//Map length
      PROT_READ|PROT_WRITE,	// Enable reading & writting to mapped memory
      MAP_SHARED,       	//Shared with other processes
      mem_fd,           	//File to map
      GPIO_BASE         	//Offset to REGs
   );

   close(mem_fd); 

   if (gpio_map == MAP_FAILED) {
      printf("mmap error %d\n", (int)gpio_map);//errno also set!
      exit(-1);
   }
   // Always use volatile pointer!
   gpio = (volatile unsigned *)gpio_map;
}

//------------------------------------------------------ SHT-11 TWI用サブルーティン

void SHT_Init(){						//初期化
	SCK_L();
	SCK_OUT();
	SDA_IN();
}

void SHT_Conreset(){					//リセット・コンディション生成
	uint8_t i;
	SDA_H();							//Connection Reset : sdaはhで固定
	for(i=0;i>9;i++){
		SCK_TRG();
	}
}

void SHT_Trans_st(){					//転送スタート・コンディション生成
	SCK_H();
	SDA_L();
	SCK_L();
	SCK_H();
	SDA_H();
	SCK_L();

	crcval = SHT_ReverseData(sht_status_reg & 0x0f);	//crc演算用レジスタを初期化
}

void SHT_CalcCRC(uint8_t data){			//CRC用演算ルーティン. 1bitずつ処理
	uint8_t i;
	for(i=0;i<8;i++){
		if((data ^ crcval) & 0b10000000){
			crcval <<= 1;
			crcval ^= 0b00110001;
		}else{
			crcval <<= 1;
		}
		data <<= 1;
	}
}

uint8_t SHT_ReverseData(uint8_t data){	//ビット反転用ルーティン(bit7..0 -> 0..7)
	uint8_t i, revdata = 0;
	for(i=0;i<8;i++){
		revdata <<= 1;
		revdata |= (data & 0x01);
		data >>= 1;
	}
	return revdata;
}

uint8_t SHT_Write(uint8_t data){		//1バイト送信 return:0=ACK 1=NAK
	uint8_t i;

	SHT_CalcCRC(data);			//先にCRC演算しておく
	
	for(i=0;i<8;i++){			//1byte data send
		if(data & 0x80){
			SDA_H();
		}else{
			SDA_L();
		}
		data <<= 1;
		SCK_TRG();
	}
	SDA_IN();					//=SDA:input
	SCK_H();
	i = ACK;
	if(SDA_GET() == NAK) i = NAK;
	SCK_L();
	
	return(i);
}

uint8_t SHT_Read(uint8_t f_ack){		//1バイト受信 + ACK or NAK送信

	uint8_t i;
	uint8_t data = 0;
	SDA_IN();
	for(i=0;i<8;i++){
		SCK_H();
		data <<= 1;
		if(SDA_GET()) data++;
		SCK_L();
	}
	if(f_ack == ACK) SDA_L();	//ACK送信
	SCK_TRG();
	SDA_IN();

	return(data);
}
uint8_t SHT_WaitforReady(void){			//計測完了までのウェイト.完了でSDAがLになる

	struct timespec pulse;
	uint8_t timeout = 30;		// *10msec = 300msでタイムアウト

	while(SDA_GET()){
		sleepMsec(10);
		if (!(--timeout)) return 1;
	}
	return 0;
}

void waitUsec (uint32_t wait){			//マイクロ秒のウェイト
   struct timespec now, pulse;
   int nsec;

	wait *= 1000;				// us->ns

	clock_gettime(CLOCK_MONOTONIC, &pulse);
	while(nsec < wait){
		clock_gettime(CLOCK_MONOTONIC, &now);
		if(now.tv_sec > pulse.tv_sec){
			nsec = 1000000000;
		}else{
			nsec = 0;
		}
		nsec += now.tv_nsec - pulse.tv_nsec;
	}
}
void sleepMsec (uint32_t waitmsec){		//ミリ秒のウェイト
	struct timespec pulse;
	clock_gettime(CLOCK_MONOTONIC, &pulse);
  	pulse.tv_nsec += (waitmsec * 1000 * 1000);	//sleep 10msec (now+=10msec)
  	if(pulse.tv_nsec >= (1000 * 1000 * 1000)){
		pulse.tv_nsec -= (1000 * 1000 * 1000);
    	pulse.tv_sec += 1;
	}
  	clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &pulse, NULL);
}

//-------------------------------------- 温度/湿度測定ルーティン

void SHT_measureValue(){

	float temp_value, hum_value,rawfl;
	uint16_t raw_data;
	uint8_t val;

	SHT_Init();
	SHT_Conreset();
	SHT_Trans_st();					//転送スタート・コンディション
	if(SHT_Write(CMD_SRES)==NAK){	//ソフトウェアリセットコマンド送信
		printf("ERR:デバイスからの応答がありません.\n");
		exit(-1);
	}
	sleepMsec(20);

// --------------- 温度測定
	SHT_Trans_st();
	if(SHT_Write(CMD_Temp)==NAK){	//温度測定コマンド送信
		printf("ERR:デバイスからの応答がありません.\n");
      	exit(-1);
	}
	if(SHT_WaitforReady()){			//測定完了までウェイト
		printf("ERR:測定中にタイムアウトしました.\n");
		exit(-1);
	}
	val = SHT_Read(ACK);			//上位バイト受信
	SHT_CalcCRC(val);
	raw_data= val << 8;

	val = SHT_Read(ACK);			//下位バイト受信
	SHT_CalcCRC(val);
	raw_data+= val;

	val = SHT_Read(ACK);			//CRCチェックバイト受信
	if (crcval != SHT_ReverseData(val)){
		printf("ERR:CRCチェックが一致しません.\n");
		exit(-1);
	}

	temp_value = (float)(raw_data) * 0.01 - 40;
	
// ------------- 湿度測定	
	SHT_Trans_st();
	if(SHT_Write(CMD_Hum)==NAK){	//湿度測定コマンド送信
		printf("ERR:デバイスからの応答がありません.\n");
		exit(-1);
	}
	if(SHT_WaitforReady()){			//測定完了までウェイト
		printf("ERR:測定中にタイムアウトしました.\n");
		exit(-1);
	}
	val = SHT_Read(ACK);			//上位バイト受信
	SHT_CalcCRC(val);
	raw_data = val << 8;

	val = SHT_Read(ACK);			//下位バイト受信
	SHT_CalcCRC(val);
	raw_data += val;

	val = SHT_Read(ACK);			//CRCチェックバイト受信
	if (crcval != SHT_ReverseData(val)){
		printf("ERR:CRCチェックが一致しません.");
		exit(-1);
	}
	//受信データをもとに実温度/実湿度に変換
	rawfl = (float)raw_data;
	hum_value = -4.0 + 0.0405 * rawfl - (0.0000028 * rawfl * rawfl);
	hum_value += (temp_value - 25) * (0.01 + 0.00008 * rawfl);

	if (hum_value>100) hum_value = 100;
	if (hum_value<0.1) hum_value = 0.1;
	
	//値を標準出力(表示)
	printf("Tmp:%0.2f, Hum:%0.2f\n",temp_value, hum_value);
	
}//SHT_measureValue

//-------------------------------------- main

int main(int argc, char **argv){
	int ret;
	map_io();
	SHT_measureValue();
	ret = munmap((void *)gpio_map, BLOCK_SIZE);
	return ret;
} // main




	